from statistics import mean
from typing import Dict, Any, List
import logging
import os
from datetime import datetime, timedelta
import calendar

from dotenv import load_dotenv
load_dotenv()
LOG_LEVEL = os.environ.get('LOG_LEVEL', 'INFO').upper()

import logging
logging.basicConfig(
    format='%(asctime)s %(levelname)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=LOG_LEVEL
)

from helpers.SimpleBeatSaverAPI import SimpleBeatSaverAPI

from saberlist.utils import load_history, save_history, prompt_for_mapper_ids

def playlist_strategy_beatsaver_acc(
    song_count: int = 40
) -> List[Dict[str, Any]]:
    """
    Build and format a list of songs based on the highest average accuracy from recent maps.
    Prompts the user for the month and year, defaulting to last month.
    Excludes any map that's in the history, regardless of difficulty.
    
    :param song_count: The number of songs to include in the playlist. Default is 40.
    :return: A list of dictionaries containing song information for the playlist.
    """
    history = load_history()
    history.setdefault('beatsaver', {})
    history.setdefault('playlist_counts', {})

    # Get last month's date
    today = datetime.now()
    last_month = today.replace(day=1) - timedelta(days=1)
    default_month = last_month.month
    default_year = last_month.year

    # Prompt for month and year
    while True:
        month_input = input(f"Enter month (1-12, default {default_month}): ").strip() or str(default_month)
        year_input = input(f"Enter year (default {default_year}): ").strip() or str(default_year)

        try:
            month = int(month_input)
            year = int(year_input)
            if 1 <= month <= 12 and 2000 <= year <= datetime.now().year:
                break
            else:
                print("Invalid month or year. Please try again.")
        except ValueError:
            print("Invalid input. Please enter numbers only.")

    # Get the current count for highest accuracy and increment it
    count_key = f"beatsaver-{year}-{month:02d}"
    current_count = history['playlist_counts'].get(count_key, 0)
    new_count = current_count + 1
    history['playlist_counts'][count_key] = new_count

    leaderboard_data = map_leaders_by_month(month=month, year=year)

    if not leaderboard_data:
        logging.error(f"No map+difficulty data available for {calendar.month_name[month]} {year}.")
        return [], ""

    # Sort the data by average_accuracy in descending order
    sorted_data = sorted(
        leaderboard_data,
        key=lambda x: x['average_accuracy'],
        reverse=True
    )

    playlist_data = []
    for entry in sorted_data:
        if len(playlist_data) >= song_count:
            break

        song_hash = entry['hash']

        # Check history to avoid reusing any map, regardless of difficulty
        if song_hash in history['beatsaver']:
            logging.debug(f"Skipping song {song_hash} as it's in history.")
            continue

        playlist_data.append({
            'hash': song_hash,
            'difficulties': entry['difficulties'],
            'songName': entry['map_name']
        })

        # Log the song addition
        difficulty = entry['difficulties'][0]['name']
        logging.info(f"Song added: {entry['map_name']} ({difficulty}) - Average Accuracy: {entry['average_accuracy'] * 100:.2f}%")

        # Update history (now we're just adding the song hash, not the difficulty)
        history['beatsaver'][song_hash] = True

    # Log if no songs were added
    if not playlist_data:
        logging.info(f"No new songs found to add to the playlist for {calendar.month_name[month]} {year} based on history.")
    else:
        logging.info(f"Total songs added to playlist: {len(playlist_data)}")

    save_history(history)

    return playlist_data, f"beatsaver-{year}-{month:02d}-{new_count:02d}"

def playlist_strategy_beatsaver_curated(
    api: SimpleBeatSaverAPI,
    song_count: int = 50
) -> List[Dict[str, Any]]:
    """
    Build a playlist from BeatSaver's curated songs list.
    
    :param api: SimpleBeatSaverAPI instance for making API calls
    :param song_count: Number of songs to include in the playlist
    :return: Tuple of (playlist data, playlist title)
    """
    history = load_history()
    history.setdefault('beatsaver_curated', {})
    history.setdefault('playlist_counts', {})

    # Get the current count and increment it
    count_key = 'beatsaver_curated'
    current_count = history['playlist_counts'].get(count_key, 0)
    new_count = current_count + 1
    history['playlist_counts'][count_key] = new_count

    curated_songs = api.get_curated_songs()
    if not curated_songs:
        raise RuntimeError("Unable to fetch curated songs from BeatSaver.")
    logging.info(f"Found {len(curated_songs)} curated songs on BeatSaver.")

    playlist_data = []
    for song in curated_songs:
        if len(playlist_data) >= song_count:
            break

        song_hash = song.get('hash')
        
        if song_hash in history['beatsaver_curated']:
            logging.debug(f"Skipping song {song_hash} as it's in history")
            continue

        playlist_data.append(song)
        logging.info(f"Song added: {song['songName']}")

        # Update history
        history['beatsaver_curated'][song_hash] = True

    if not playlist_data:
        logging.info("No new songs found to add to the playlist based on history")
    else:
        logging.info(f"Total songs added to playlist: {len(playlist_data)}")

    save_history(history)
    playlist_title = f"curated-{new_count:02d}"

    return playlist_data, playlist_title

def map_leaders_by_month(month: int = 9, year: int = 2024, game_modes: List[str] = ['Standard']) -> List[Dict]:
    """
    Gathers a month's worth of maps using the BeatSaver latest maps endpoint,
    prioritizes map difficulties where players have already set good scores,
    and calculates the average accuracy for each map+difficulty.

    Args:
        month: The month to gather maps for.
        year: The year to gather maps for.
        game_modes: The game modes to include.

    Returns:
        A list of dictionaries, each containing:
            - hash: Hash of the map
            - difficulties: List of difficulties with their characteristics
            - map_name: Name of the map
            - average_accuracy: Average accuracy of the leaderboard
    """
    beatleader_api = SimpleBeatLeaderAPI(cache_expiry_days=30)
    beatsaver_api = BeatSaverAPI(cache_expiry_days=30)

    logging.debug(f"Fetching maps for {month}/{year}")
    map_data = beatsaver_api.get_maps(year=year, month=month, page_size=100)

    collected_data = []

    for i, map_entry in enumerate(map_data):
        
        # Ensure there are versions available
        if not map_entry.versions:
            logging.warning(f"No versions found for map: {map_entry.name}")
            continue

        latest_version = max(map_entry.versions, key=lambda version: version.created_at)
        song_hash = latest_version.hash_

        for diff in latest_version.diffs:
            if diff.characteristic not in game_modes:
                continue

            logging.info(f"Getting leaderboard for {i+1}/{len(map_data)}: {map_entry.name}")
            leaderboard_data = beatleader_api.get_leaderboard(song_hash, diff.difficulty)

            if not leaderboard_data:
                logging.debug(f"No leaderboard data for {map_entry.name} [{diff.difficulty}], skipping")
                continue

            # Calculate average accuracy
            accuracies = [entry.get('accuracy', 0) for entry in leaderboard_data if 'accuracy' in entry]
            if not accuracies:
                logging.debug(f"No accuracy data for {map_entry.name} [{diff.difficulty}]")
                continue

            avg_accuracy = mean(accuracies)

            collected_data.append({
                'hash': song_hash,
                'difficulties': [
                    {
                        'name': diff.difficulty,
                        'characteristic': diff.characteristic
                    }
                ],
                'map_name': map_entry.name,
                'average_accuracy': avg_accuracy
            })

    logging.info(f"Collected leaderboards for {len(collected_data)} map+difficulty combinations, orderable by average accuracy of top ten plays for {month}/{year}.")
    return collected_data

def playlist_strategy_beatsaver_mappers(
    api: SimpleBeatSaverAPI,
    song_count: int = 50
) -> List[Dict[str, Any]]:
    """
    Build a playlist from maps created by specified mappers on BeatSaver,
    interleaving maps from different mappers for variety.
    
    :param api: SimpleBeatSaverAPI instance for making API calls
    :param song_count: Number of songs to include in the playlist
    :return: Tuple of (playlist data, playlist title)
    """
    history = load_history()
    history.setdefault('beatsaver_mappers', {})
    history.setdefault('playlist_counts', {})
    mapper_ids = prompt_for_mapper_ids()

    # Get the current count and increment it
    count_key = 'beatsaver_mappers'
    current_count = history['playlist_counts'].get(count_key, 0)
    new_count = current_count + 1
    history['playlist_counts'][count_key] = new_count

    # Collect maps by mapper
    maps_by_mapper = {}
    for mapper_id in mapper_ids:
        logging.info(f"Fetching maps for mapper ID: {mapper_id}")
        mapper_maps = api.get_mapper_maps(mapper_id=mapper_id, use_cache=False)
        
        if not mapper_maps:
            logging.warning(f"No maps found for mapper ID: {mapper_id}")
            continue
            
        logging.info(f"Found {len(mapper_maps)} maps from mapper ID: {mapper_id}")

        # Filter out maps that are in history
        eligible_maps = [
            song for song in mapper_maps 
            if song.get('hash') not in history['beatsaver_mappers']
        ]
        
        if eligible_maps:
            maps_by_mapper[mapper_id] = eligible_maps

    # Interleave maps from different mappers
    playlist_data = []
    while maps_by_mapper and len(playlist_data) < song_count:
        for mapper_id in list(maps_by_mapper.keys()):
            if not maps_by_mapper[mapper_id]:
                del maps_by_mapper[mapper_id]
                continue
                
            song = maps_by_mapper[mapper_id].pop(0)
            song_hash = song.get('hash')
            
            playlist_data.append(song)
            logging.info(f"Song added: {song['songName']} by mapper {mapper_id}")
            
            # Update history
            history['beatsaver_mappers'][song_hash] = True
            
            if len(playlist_data) >= song_count:
                break

    if not playlist_data:
        logging.info("No new songs found to add to the playlist based on history")
    else:
        logging.info(f"Total songs added to playlist: {len(playlist_data)}")

    save_history(history)
    playlist_title = f"mappers-{new_count:02d}"

    return playlist_data, playlist_title
