from typing import Dict, Any, List
import argparse
import logging
import os
from dotenv import load_dotenv
load_dotenv()
LOG_LEVEL = os.environ.get('LOG_LEVEL', 'INFO').upper()

import logging
logging.basicConfig(
    format='%(asctime)s %(levelname)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=LOG_LEVEL
)

from helpers.ScoreSaberAPI import ScoreSaberAPI
from helpers.BeatLeaderAPI import BeatLeaderAPI

from saberlist.utils import prompt_for_player_id, load_history, save_history, normalize_difficulty_name

def playlist_strategy_beatleader_lowest_pp(
    api: BeatLeaderAPI,
    song_count: int = 20
) -> List[Dict[str, Any]]:
    player_id = prompt_for_player_id()
    history = load_history()
    history.setdefault('beatleader_lowest_pp', {})
    history.setdefault('playlist_counts', {})

    # Get the current count for BeatLeader lowest PP and increment it
    count_key = 'beatleader_lowest_pp'
    current_count = history['playlist_counts'].get(count_key, 0)
    new_count = current_count + 1
    history['playlist_counts'][count_key] = new_count

    scores_data = api.get_player_scores(player_id)
    all_scores = scores_data.get('playerScores', [])
    if not all_scores:
        logging.warning(f"No scores found for player ID {player_id} on BeatLeader.")
        return [], ""
    logging.debug(f"Found {len(all_scores)} scores for player ID {player_id} on BeatLeader.")

    # Filter out scores with zero PP and sort by PP in ascending order
    ranked_scores = [s for s in all_scores if s.get('score', {}).get('pp', 0) > 0]
    ranked_scores.sort(key=lambda x: x.get('score', {}).get('pp', float('inf')))

    playlist_data = []
    for score_entry in ranked_scores:
        if len(playlist_data) >= song_count:
            break  # Stop if we've reached the desired number of songs

        score = score_entry.get('score', {})
        leaderboard = score_entry.get('leaderboard', {})
        
        song_hash = leaderboard.get('songHash')
        difficulty_raw = int(leaderboard.get('difficulty', ''))
        game_mode = leaderboard.get('modeName', 'Standard')
        pp = score.get('pp', 0)

        if not song_hash or not difficulty_raw:
            logging.debug(f"Skipping score due to missing song_hash or difficulty_raw: {score_entry}")
            continue

        difficulty = normalize_difficulty_name(difficulty_raw)

        # avoid reusing song+difficulty
        if song_hash in history['beatleader_lowest_pp'] and difficulty in history['beatleader_lowest_pp'][song_hash]:
            logging.debug(f"Skipping song {song_hash} with difficulty {difficulty} as it's in history.")
            continue  # Skip if already used

        # Format the song data for PlaylistBuilder
        song_dict = {
            'hash': song_hash,
            'difficulties': [
                {
                    'name': difficulty,
                    'characteristic': game_mode
                }
            ]
        }

        # Add the song to the playlist
        playlist_data.append(song_dict)
        logging.debug(f"Selected song for playlist: Hash={song_hash}, Difficulty={difficulty}, PP={pp:.2f}")

        # Update history
        history['beatleader_lowest_pp'].setdefault(song_hash, []).append(difficulty)

    # Log the final playlist
    if not playlist_data:
        logging.info("No new songs found to add to the playlist based on history for BeatLeader lowest PP.")
    else:
        for song in playlist_data:
            song_hash = song['hash']
            difficulty = song['difficulties'][0]['name']
            logging.info(f"Song added: Hash={song_hash}, Difficulty={difficulty}")
        logging.info(f"Total songs added to playlist from BeatLeader lowest PP: {len(playlist_data)}")

    save_history(history)

    return playlist_data, f"beatleader_lowest_pp-{new_count:02d}"

def playlist_strategy_scoresaber_lowest_pp(
    api: ScoreSaberAPI,
    song_count: int = 20
) -> List[Dict[str, Any]]:
    """Build and format a list of songs based on lowest PP scores from ScoreSaber, avoiding reusing the same song+difficulty."""
    
    player_id = prompt_for_player_id()
    history = load_history()
    history.setdefault('scoresaber_lowest_pp', {})
    history.setdefault('playlist_counts', {})

    # Get the current count for ScoreSaber lowest PP and increment it
    count_key = 'scoresaber_lowest_pp'
    current_count = history['playlist_counts'].get(count_key, 0)
    new_count = current_count + 1
    history['playlist_counts'][count_key] = new_count

    scores_data = api.get_player_scores(player_id, use_cache=True)
    all_scores = scores_data.get('playerScores', [])
    if not all_scores:
        logging.warning(f"No scores found for player ID {player_id}.")
        return [], ""
    logging.debug(f"Found {len(all_scores)} scores for player ID {player_id}.")
    
    # Filter out scores with zero PP and sort by PP in ascending order
    ranked_scores = [s for s in all_scores if s['score'].get('pp', 0) > 0]
    ranked_scores.sort(key=lambda x: x['score'].get('pp', float('inf')))
    
    playlist_data = []
    
    for score in ranked_scores:
        leaderboard = score.get('leaderboard', {})
        song_id = leaderboard.get('songHash')
        difficulty_raw = leaderboard.get('difficulty', {}).get('difficultyRaw', '')
        
        if not song_id or not difficulty_raw:
            logging.debug(f"Skipping score due to missing song_id or difficulty_raw: {score}")
            continue  # Skip if essential data is missing
        
        # Normalize the difficulty name
        difficulty = normalize_difficulty_name(difficulty_raw)
        game_mode = leaderboard.get('difficulty', {}).get('gameMode', 'Standard')
        if 'Standard' in game_mode:
            game_mode = 'Standard'
        
        # Check history to avoid reusing song+difficulty
        if song_id in history['scoresaber_lowest_pp'] and difficulty in history['scoresaber_lowest_pp'][song_id]:
            logging.debug(f"Skipping song {song_id} with difficulty {difficulty} as it's in history.")
            continue  # Skip if already used
        
        # Format the song data as expected by PlaylistBuilder
        song_dict = {
            'hash': song_id,
            'songName': leaderboard.get('songName', 'Unknown'),
            'difficulties': [
                {
                    'name': difficulty,
                    'characteristic': game_mode
                }
            ]
        }
        
        # Add the song to the playlist
        playlist_data.append(song_dict)
        pp_score = score['score'].get('pp', 0)
        logging.info(f"Song added: {song_dict['songName']} ({difficulty}), PP: {pp_score:.2f}")
        
        # Check if the desired number of songs has been reached
        if len(playlist_data) >= song_count:
            logging.debug(f"Reached the desired song count: {song_count}.")
            break
    
    # Log if no songs were added
    if not playlist_data:
        logging.info("No new songs found to add to the playlist based on history.")
    else:
        logging.info(f"Total songs added to playlist: {len(playlist_data)}")
    
    # Update history to avoid reusing the same song+difficulty
    for song in playlist_data:
        song_id = song['hash']
        difficulty_name = song['difficulties'][0]['name']
        history['scoresaber_lowest_pp'].setdefault(song_id, []).append(difficulty_name)
    save_history(history)
    
    return playlist_data, f"scoresaber_lowest_pp-{new_count:02d}"
